import pygame
from pygame.constants import *

from OpenGL.GL import *
from OpenGL.GLU import *

class Screen:
    
    def __init__(self,width,height):
        self.width = width
        self.height = height
        self.screen = None
        self.allGroup = pygame.sprite.Group()
        self.tracking = None
        
    def realize(self):
        sizeTuple = (self.width, self.height)
        self.screen = pygame.display.set_mode( sizeTuple,
                                               DOUBLEBUF )
   
    def paint(self):
        self.screen.fill( (0,0,0) )
        
        pygame.display.flip()

class GLScreen(Screen):
    
    FAR_CULLED = 5000
    NEAR_CULLED = 1
    CAMERA_TOPDOWN = 1
    CAMERA_CHASE = 2
    CAMERA_COOLSCREENSHOT = 3
    
    def __init__(self,width,height):
        Screen.__init__(self,width,height)
        self.sprites = []
        self.hudSprites = []
        if self.height == 0:
            self.height = 1
        self.zoom = 1000
        self.cameraMode = self.CAMERA_TOPDOWN
        self.loadedTexes = {}
        
    def add(self,sprite):
        self.sprites.append(sprite)
        
    def addToHud(self,sprite):
        self.hudSprites.append(sprite)
        
    def clear(self):
        self.sprites = []
        self.hudSprites = []
        
    def load(self,filename):
        # Thanks to http://www.pygame.org/wiki/SimpleOpenGL2dClasses?parent=CookBook
        
        if filename in self.loadedTexes:
            return self.loadedTexes[filename]
            
        textureSurface = pygame.image.load(filename)
        textureData = pygame.image.tostring(textureSurface, "RGBA", 1)
        width = textureSurface.get_width()
        height = textureSurface.get_height()
     
        texture = glGenTextures(1)
        glBindTexture(GL_TEXTURE_2D, texture)
        #print filename, "bound to texture",texture
        glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR)
        glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR)
        glTexImage2D(GL_TEXTURE_2D, 0, GL_RGBA, width, height, 0, GL_RGBA,
            GL_UNSIGNED_BYTE, textureData)
     
        self.loadedTexes[filename] = (texture,width,height)
     
        return texture, width, height
        
    def realize(self):
        sizeTuple = (self.width, self.height)
        self.screen = pygame.display.set_mode( sizeTuple,
                                               DOUBLEBUF | OPENGL)
        
        glEnable(GL_TEXTURE_2D)
        glEnable(GL_NORMALIZE)
        glViewport(0,0,self.width,self.height)
        
    def paint(self):
        glEnable(GL_DEPTH_TEST)
        glEnable(GL_LIGHTING)
        glMatrixMode(GL_PROJECTION)
        glLoadIdentity()
        gluPerspective(60,self.width/self.height,
                        self.NEAR_CULLED,self.FAR_CULLED)
        glMatrixMode(GL_MODELVIEW)
        glClear(GL_COLOR_BUFFER_BIT|GL_DEPTH_BUFFER_BIT)
        glLoadIdentity()
        
        if self.tracking:
            if self.cameraMode == self.CAMERA_TOPDOWN:
                gluLookAt(self.tracking.x,self.tracking.y, self.zoom,
                          self.tracking.x,self.tracking.y,1, 0,1,0)
            elif self.cameraMode == self.CAMERA_CHASE:
                backwards = -(self.tracking.velocity.unit())
                backwards *= self.zoom / 10
                gluLookAt(self.tracking.x + backwards.x,
                          self.tracking.y + backwards.y,
                          7,
                          self.tracking.x,self.tracking.y,1,
                          0,0,1)
            elif self.cameraMode == self.CAMERA_COOLSCREENSHOT:
                forwards = self.tracking.velocity.unit()
                forwards *= self.zoom / 10
                gluLookAt(self.tracking.x + forwards.x,
                           self.tracking.y + forwards.y,
                           10,
                           self.tracking.x,self.tracking.y,1,
                           0,0.5,0.5)
        else:
            gluLookAt(0,0,self.zoom, 0,0,0, 0,1,0)
        
        glEnable(GL_LIGHT0)
        glLightfv(GL_LIGHT0, GL_POSITION, [200.0, 200.0, 1000, 0.0 ])
        glLightfv(GL_LIGHT0, GL_AMBIENT, [ 0.1, 0.1, 0.1, 1.0])
        glLightfv(GL_LIGHT0, GL_DIFFUSE, [ 0.8, 0.8, 0.8, 1.0])
        glLightfv(GL_LIGHT0, GL_SPECULAR, [1.0, 1.0, 1.0, 1.0])
        
        if self.tracking:
            x,y,z = self.tracking.velocity.unit()
        
            glEnable(GL_LIGHT1)
            glLightfv(GL_LIGHT1, GL_POSITION, [self.tracking.x,
                       self.tracking.y, 5, 1.0])
            glLightfv(GL_LIGHT1, GL_DIFFUSE, [ 0.9, 0.9, 0.5])
            glLightfv(GL_LIGHT1, GL_SPOT_CUTOFF, 30.0)
            glLightfv(GL_LIGHT1, GL_SPOT_DIRECTION, (x,y,z))
        
        for sprite in self.sprites:
            sprite.paint(self)
            
        self.paintHud()
            
        glFlush()
        pygame.display.flip()
        
    def paintHud(self):
        glDisable(GL_DEPTH_TEST)
        glDisable(GL_LIGHTING)
        glMatrixMode(GL_PROJECTION)
        glLoadIdentity()
        gluOrtho2D(0, self.width, self.height, 0)
        
        glMatrixMode(GL_MODELVIEW)
        glLoadIdentity()
        
        for sprite in self.hudSprites:
            sprite.paint(self)
        
    def toggleCameraMode(self):
        if self.cameraMode == self.CAMERA_TOPDOWN:
            self.cameraMode = self.CAMERA_CHASE
        elif self.cameraMode == self.CAMERA_CHASE:
            self.cameraMode = self.CAMERA_COOLSCREENSHOT
        elif self.cameraMode == self.CAMERA_COOLSCREENSHOT:
            self.cameraMode = self.CAMERA_TOPDOWN
        
    def zoomIn(self,amount=100):
        self.zoomOut(-amount)
        
    def zoomOut(self,amount=100):
        self.zoom += amount
        if self.zoom < self.NEAR_CULLED:
            self.zoom = self.NEAR_CULLED + 1
        if self.zoom > self.FAR_CULLED:
            self.zoom = self.FAR_CULLED - 1

class Engine:
    
    def __init__(self,screen):
        self.screen = screen
        self.running = False
        self.updateables = []
        
    def add(self,obj,addToScreenToo = True):
        if addToScreenToo: self.screen.add(obj)
        self.updateables.append(obj)
        
    def keyEvent(self,event,pressed):
        pass
       
    def mouseButton(self,event,pressed):
        pass
        
    def run(self):
        self.running = True
        
        clock = pygame.time.Clock()
        clock.tick()
        
        while(self.running):
            event = pygame.event.poll()
            while(event.type != NOEVENT):
                if event.type == QUIT:
                    self.running = False
                    break
                elif (event.type == MOUSEBUTTONDOWN or
                       event.type ==MOUSEBUTTONUP):
                    self.mouseButton(event,event.type == MOUSEBUTTONDOWN)
                    pass
                elif (event.type == KEYDOWN or
                       event.type== KEYUP):
                    self.keyEvent(event,event.type==KEYDOWN)
                    pass
                event = pygame.event.poll()
            self.screen.paint()
            
            curTime = clock.tick()
            elapsed = curTime / 1000.0
            
            for updateable in self.updateables:
                updateable.update(elapsed)
            pygame.time.delay(40)