import common
import os
import math
import random
import operator

from OpenGL.GL import *
from OpenGL.GLU import *

from vector import Vector
import importer
import font

class Sprite:
    
    def __init__(self):
        pass
    
    def paint(self,screen):
        pass
    
    def update(self,delay):
        pass
    
class GLSprite(Sprite):
    
    def __init__(self):
        Sprite.__init__(self)
    
    def paint(self,screen=None):
        Sprite.paint(self,screen)
    
class Label(GLSprite):
    
    def __init__(self,text,font,x=0,y=0):
        GLSprite.__init__(self)
        self.text = text
        self.font = font
        self.x = x
        self.y = y
        
    def paint(self,screen):
        self.font.glRender(self.x,self.y,self.text)
        
class Image(GLSprite):
    
    def __init__(self,filename,x=0, y=0):
        GLSprite.__init__(self)
        self.filename = filename
        self.texture = None
        self.x = x
        self.y = y
        
    def paint(self,screen):
        if not self.texture:
            self.texture,w,h = screen.load(self.filename)
            self.w = w
            self.h = h
            
        glColor(1.0, 1.0, 1.0)
        glBindTexture(GL_TEXTURE_2D, self.texture)
        glPushMatrix()
        
        glTranslate(self.x,self.y,0)
        
        glBegin(GL_QUADS)
        glTexCoord(0,0); glVertex(0,self.h)
        glTexCoord(1,0); glVertex(self.w, self.h)
        glTexCoord(1,1); glVertex(self.w, 0)
        glTexCoord(0,1); glVertex(0, 0)
        glEnd()
        
        glPopMatrix()
        
class Hud(GLSprite):
    
    def __init__(self,city,car):
        GLSprite.__init__(self)
        self.city = city
        self.car = car
       
        filename = os.path.join("font","freesansbold.ttf")
        self.bigfont = font.GLFont(os.path.join(filename),72)
        self.smallfont = font.GLFont(os.path.join(filename),14)
        self.damageLabel = Label("Damage",self.smallfont,20,10)
        self.gameOverLabel = Label(" ",self.bigfont,150,120)
        self.instructions = Label(
            "Press <escape> to exit\nPress <space> to start over",
            self.smallfont,0, 400)
        self.gameIsOver = False
        
    def loadTextures(self,screen):
        pass
        
    def gameOver(self,message):
        self.gameOverLabel.text = message
        self.gameIsOver = True
        
    def paint(self,screen=None):
        
        if self.car.dispatcher.robberBeingTailed():
            color = (0.8, 0.1, 0.1)
        else:
            color = (0.1, 0.1, 0.8)
            
        glColor(color)
        
        glBindTexture(GL_TEXTURE_2D, 0)
        glBegin(GL_QUADS)
        glVertex(0, 0)
        glVertex(0, screen.height / 10)
        glVertex(screen.width, screen.height / 10)
        glVertex(screen.width, 0)
        glEnd()
        
        glColor(1.0, 1.0, 1.0)
        glEnable(GL_BLEND)
        glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA)
        
        self.damageLabel.paint(screen)
        if self.gameIsOver:
            glColor(0.8, 0.05, 0.05)
            self.gameOverLabel.paint(screen)
            glColor(0.0, 0.5, 0.0)
            self.instructions.paint(screen)
        
        glDisable(GL_BLEND)
        glColor(0.0, 0.0, 0.0)
        # Outline damage
        right = 5 + screen.width / 3
        glBindTexture(GL_TEXTURE_2D, 0)
        glBegin(GL_QUADS)
        glVertex(5, 30); glVertex(5, 45)
        glVertex(right, 45)
        glVertex(right, 30)
        glEnd()
        
        percentage = float(self.car.damage) / float(self.car.maxDamage)
        glColor( percentage, 1-percentage, 0.0)
        right = 6 + (right - 2) * percentage
        glBegin(GL_QUADS)
        glVertex(6, 31); glVertex(6, 44)
        glVertex(right, 44); glVertex(right, 31)
        glEnd()

    
class Car(GLSprite):
    
    def __init__(self,city,model,loader,scale=1.0):
        GLSprite.__init__(self)
        self.location = Vector(0.0,0.0)
        self.velocity = Vector(0.0,0.0)
        self.heading = 0
        self.dHeading = 0
        self.accel = 0
        self.zeroToInOneSecond = 5
        self.city = city
        self.weight = 1
        self.cornering = 0.25
        self.cornerAngle = 10
        self.offsetFromY = 0
        self.offsetFromX = 0
        i = importer.OBJImporter(loader)
        self.model = i.process(os.path.join('models',model),scale)
        self.topSpeed = 5
        self.radius = 20
        self.scale = scale
        self.dispatcher = None
        
    def __getattr__(self,attr):
        if attr == 'x':
            return self.__dict__['location'].x
        elif attr == 'y':
            return self.__dict__['location'].y
        raise AttributeError("No %s in Car" % attr)
    
    def accelerate(self,pressed):
        if pressed:
            self.accel = 1
        else:
            self.accel = 0
    
    def bounce(self, with = None):
        if(with):
            diff = self.velocity - with.velocity
            self.velocity -= diff
            with.velocity -= diff
        else:
            self.velocity = -( self.velocity * 1.5)
        #self.accel = 0
    
    def collideDetectWithCity(self):
        tX,tY = self.city.tileFromWorld(self.location.x,
                                        self.location.y)
        collide = False
        if self.city.outOfBounds(tX,tY):
            collide = True
        else:
            type = self.city.at(tX,tY)
            if not self.city.isRoad(type):
                collide = True
                
        if collide:
            self.bounce()
    
    def collideCheck(self, other):
        loc = self.location + (self.offsetFromY * math.sin(self.heading))
        otherLoc = other.location + (other.offsetFromY * math.sin(other.heading))
       
        if (loc.dist(otherLoc) <= self.radius):
            #print "%s collided with %s" % (self,other)
            
            self.bounce(other)
            #other.bounce(self)
    
    def decelerate(self,pressed):
        if pressed:
            self.accel = -1.0
        else:
            self.accel = 0
    
    def paint(self,screen=None):
        glPushMatrix()
        glTranslate(self.x,self.y,1)
        glRotate(self.heading,0,0,1)
        glColor3f(1.0,0.0,0.0)

        glCallList(self.model)
        glPopMatrix()
    
    def turnLeft(self,pressed):
        if pressed:
            self.dHeading = self.cornerAngle
        else:
            self.dHeading = 0
        
    def turnRight(self,pressed):
        if pressed:
            self.dHeading = -self.cornerAngle
        else:
            self.dHeading = 0
    
    def speed(self):
        return self.velocity.magnitude()
    
    def update(self,delay):
        
        self.heading += self.dHeading
                
        accelVector = Vector(0,0)
        
        self.collideDetectWithCity()
        
        # Are we accelerating?  
        if self.accel:
            # From 'Physics for Game Developers' O'Reilly, p30
            deltaSpeed = (delay/2)*self.zeroToInOneSecond * self.accel
            
            accelVector = Vector(deltaSpeed,self.heading,0,False)
            self.velocity += accelVector
            
        # Drag
        coeff = 0.015 # From O'Reilly, p169
        coeff = coeff * self.weight
        
        speed = self.velocity.magnitude()
        if speed:
            # Move part of our velocity to our heading
            headVector = Vector(speed,self.heading,0,False)
            velVector = self.velocity
            
            if headVector.unit() != velVector.unit():
                #print "Velocity before:",self.velocity,
                self.velocity = ((headVector * self.cornering) +
                                 (velVector * (1-self.cornering)))
            accel = (self.velocity - coeff)/self.weight
           
            # Enforce speed limit
            accel = accel.clampMagnitude(self.topSpeed)
            
            self.location += accel
    
class PlayerCar(Car):
    
    def __init__(self,city,loader):
        Car.__init__(self,city,'Nova Car',loader,5)
        self.offsetFromY = -2 * self.scale
        self.topSpeed = 6
        self.damage = 0
        self.maxDamage = 100
        self.engine = None
        self.hud = None
        self.didBounce = 0
    
    def bounce(self, with=None):
        oldSpeed = self.velocity.magnitude()
        Car.bounce(self,with)
        self.didBounce += 1
    
    def collideCheck(self, other):
        #dist = self.location.dist(other.location)
        #if dist < 50:
        #    print "player checking collide with",other,"dist is",dist
        return Car.collideCheck(self,other)
    
    def collideDetectWithCity(self):
        tX,tY = self.city.tileFromWorld(self.location.x,
                                        self.location.y)
                                        
        Car.collideDetectWithCity(self)
        
        if(self.city.outOfBounds(tX,tY)): return
        
        ontile = self.city.at(tX,tY)
        offsets = ([tX+1,tY], [tX-1,tY],
                    [tX,tY+1], [tX,tY-1])
        for item in offsets:
            if not self.city.outOfBounds(*item):
                at = self.city.at(*item)
                if (at == self.city.SAFEHOUSE and
                     not self.dispatcher.robberBeingTailed() and
                     ontile & self.city.ROAD_MASK):
                    self.gameover("Victory!")
    
    def gameover(self,message):
        if self.engine:
            self.engine.gameOver()
            self.hud.gameOver(message)
        else:
            print "***",message
            raise SystemExit
    
    def paint(self,screen=None):
        glPushMatrix()
        glTranslate(self.x,self.y,self.scale)
        glRotate(self.heading+90,0,0,1)
        glTranslate(0,-2 * self.scale,0)
        
        glMaterialfv(GL_FRONT_AND_BACK, GL_AMBIENT, [0.1,0.1,0.1,1.0])
        glMaterialfv(GL_FRONT_AND_BACK, GL_DIFFUSE, [1.0,0.0,0.0,1.0])
        glMaterialfv(GL_FRONT_AND_BACK, GL_SPECULAR, [0.0,0.0,0.0,1.0])
        glColor(1.0,1.0,1.0,1.0)
        
        #glBindTexture(GL_TEXTURE_2D, 0)
        glCallList(self.model)
        glPopMatrix()
        
    def update(self,delay):
        Car.update(self,delay)
        
        if(self.didBounce):
            self.damage += self.didBounce * delay * 10
            if self.damage > self.maxDamage:
                self.gameover("Crashed!")
            self.didBounce = 0
        
        
class PoliceCar(Car):
    
    MODE_PATROL = 1
    MODE_CHASE = 2
    MODE_TRAIL = 3
    
    def __init__(self,city, criminal,loader):
        Car.__init__(self,city,'kia rio',loader,2)
        self.chosenHeading = None
        self.cornering = 0.9
        self.immediateWaypoint = None
        self.cornerAngle = 20
        self.criminal = criminal
        self.mode = self.MODE_PATROL
        self.lastLoc = [-1,-1]
        self.sight = 5
        self.offsetFromY = -2 * self.scale
        
    def checkForModeTransition(self,tX,tY):
        # Do we have line-of-sight?
        carLoc = self.criminal.location
        carX,carY = self.city.tileFromWorld(carLoc.x,carLoc.y)
        dx = tX - carX
        dy = tY - carY
        lineOfSight = ((dx == 0 and abs(dy) <= self.sight) or
                       (dy == 0 and abs(dx) <= self.sight))
        
        if self.mode == self.MODE_PATROL and lineOfSight:
            self.mode = self.MODE_CHASE
            self.immediateWaypoint = None
            
        # TODO: Change to MODE_TRAIL
        if self.mode == self.MODE_CHASE and (not lineOfSight):
            self.mode = self.MODE_PATROL
        
    def chooseChaseWaypoint(self,tX,tY):
        carLoc = self.criminal.location
        carX,carY = self.city.tileFromWorld(carLoc.x,carLoc.y)
        dx = common.sign(carX -tX)
        dy = common.sign(carY -tY)

        exits = self.city.exitsFromRoad(tX,tY)
        if [dx,dy] in exits:
            targetX,targetY = tX+dx, tY +dy
            worldX,worldY = self.city.tileToWorldCentered(
                targetX,targetY )
            self.lastLoc = tX,tY
            return Vector(worldX,worldY)
        else:
            # Can't find it, use patrol behavior
            return self.choosePatrolWaypoint(tX,tY)
        
    def choosePatrolWaypoint(self,tX,tY):
           
        targets = self.city.exitsFromRoad(tX,tY)
        
        targetX,targetY = self.lastLoc
        while(targetX == self.lastLoc[0] and
               targetY == self.lastLoc[1] and
               len(targets) > 1):
            chosenIndex = random.randrange(len(targets))
            target = targets[chosenIndex]
        
            targetX,targetY = tX+target[0], tY + target[1]
        
        worldX,worldY = self.city.tileToWorldCentered(
            targetX,targetY )
        self.lastLoc = tX,tY
        return Vector(worldX,worldY)
    
    def paint(self,screen=None):
        glBindTexture(GL_TEXTURE_2D, 0)
        glPushMatrix()
        glTranslate(self.x,self.y,5)
        glRotate(self.heading+90,0,0,1)
        glRotate(90, 1, 0, 0)
        glRotate(90, 0, 1, 0)
        glTranslate(0,-4 * self.scale,0)
        #glColor3f(0.0,0.0,1.0)
        
        glMaterialfv(GL_FRONT_AND_BACK, GL_AMBIENT, [0.1,0.1,0.1,1.0])
        glMaterialfv(GL_FRONT_AND_BACK, GL_DIFFUSE, [0.0,0.0,1.0,1.0])
        glMaterialfv(GL_FRONT_AND_BACK, GL_SPECULAR, [0.0,0.0,0.0,1.0])
        
        glCallList(self.model)
        glPopMatrix()
        
#        if self.immediateWaypoint:
#        
#            glPushMatrix()
#            glMaterialfv(GL_FRONT_AND_BACK, GL_AMBIENT, [0.1,0.1,0.1,1.0])
#            glMaterialfv(GL_FRONT_AND_BACK, GL_DIFFUSE, [0.0,0.0,1.0,1.0])
#            glMaterialfv(GL_FRONT_AND_BACK, GL_SPECULAR, [1.0, 1.0, 0.0, 1.0])
#            
#            glTranslate(self.immediateWaypoint.x,
#                         self.immediateWaypoint.y, 10)
#            glBegin(GL_QUADS)
#            glVertex(-5, -5, 0)
#            glVertex(-5,  5, 0)
#            glVertex( 5,  5, 0)
#            glVertex( 5, -5, 0)
#            glEnd()
#            glPopMatrix()
    
    def turnLeft(self,amount):
        self.dHeading = amount
        
    def turnRight(self,amount):
        self.dHeading = -amount
    
    def update(self,delay):
        
        heading = self.heading % 360
        while(heading < 0):
            heading += 360
           
        tX,tY = self.city.tileFromWorld(self.location.x,
                                        self.location.y)
        
        self.checkForModeTransition(tX,tY)
        
        chooseNewWaypoint = False
        
        if self.immediateWaypoint != None:
            dist = self.location.dist(self.immediateWaypoint)
                                              
            if(dist < 10):
                chooseNewWaypoint = True
        else:
            chooseNewWaypoint = True
            
        if(chooseNewWaypoint):   
            if self.mode == self.MODE_PATROL:
                self.immediateWaypoint = self.choosePatrolWaypoint(tX,tY)
            elif self.mode == self.MODE_CHASE:
                self.immediateWaypoint = self.chooseChaseWaypoint(tX,tY)
        
        headVector = Vector(1.0, heading, 0, False)
        wantVector = self.immediateWaypoint - self.location
        
        cross = headVector.cross(wantVector)

        between = math.degrees(headVector.angleBetween(wantVector))
        between = between > 20 and 20 or between
        if between > 10:
            if cross.z < 0:
                self.turnRight(between)
            elif cross.z > 0:
                self.turnLeft(between)
            else:
                self.turnLeft(20)
        else:
            self.turnRight(0)
        
        self.accelerate(1)
            
        Car.update(self,delay)
        
#        print self.city.tileFromWorld(self.location.x,
#                                       self.location.y)
    
class Ground(GLSprite):
    
    ORDINARY = 0
    SAFEHOUSE = 1
    
    def __init__(self,city):
        GLSprite.__init__(self)
        self.city = city
        self.textures = []
        self.buildTextures = []
        self.groundList = None
        
    def loadTextures(self,screen):
        
        for index in range(self.city.ROADS):
            filename = os.path.join("images","road_%d.png" % index)
            tex,width,height = screen.load(filename)
            self.textures.append(tex)
            
        grass = os.path.join("images","grass.png")
        tex,width,height = screen.load(grass)
        #print "Loaded a texture!",tex,width,height
        self.textures.append(tex)
        
        for index in range(1,8):
            filename = os.path.join("images","building_%d.png" % index)
            tex,width,height = screen.load(filename)
            self.buildTextures.append(tex)
            
        self.safeTextures = []
        filename = os.path.join("images", "brick_wall.png")
        tex,width,height = screen.load(filename)
        self.safeTextures.append(tex)
        
        filename = os.path.join("images", "brick_door.png")
        tex,width,height = screen.load(filename)
        self.safeTextures.append(tex)
        
    def paint(self,screen=None):
        
        if(not self.textures):
            self.loadTextures(screen)
        
        if self.groundList == None:
            self.groundList = self.paintGround()
            
        glCallList(self.groundList)
        
    def paintGround(self):
        id = glGenLists(1)
        glNewList(id,GL_COMPILE)
    
        width = self.city.width * common.BLOCK_UNITS
        height = self.city.height * common.BLOCK_UNITS
        upper = -(height / 2.0)
        left = -(width/2.0)
        
        rowNum = 0
        for row in self.city.rows():
            colNum = 0
            for col in row:
                #glColor3f(1.0,1.0,1.0)
                if col > self.city.ROADS:
                    texIndex = self.city.ROADS
                else:
                    texIndex = col
                glBindTexture(GL_TEXTURE_2D,self.textures[texIndex])
                
                glMaterialfv(GL_FRONT_AND_BACK, GL_DIFFUSE, [0.8,0.8,0.8,1.0])
                if col & self.city.ROAD_MASK:
                    glMaterialfv(GL_FRONT_AND_BACK, GL_SPECULAR, [1.0, 1.0, 1.0,1.0])
                else:
                    glMaterialfv(GL_FRONT_AND_BACK, GL_SPECULAR, [0.0,0.0, 0.0, 1.0])
                
                glBegin(GL_QUADS)
                x1,y1 = self.city.tileToWorld(colNum,rowNum)
                x2,y2 = self.city.tileToWorld(colNum+1,rowNum+1)
                
                glTexCoord2f(0,1); glNormal(0,0,1); glVertex(x1,y1,0)
                glTexCoord2f(0,0); glNormal(0,0,1); glVertex(x1,y2,0)
                glTexCoord2f(1,0); glNormal(0,0,1); glVertex(x2,y2,0)
                glTexCoord2f(1,1); glNormal(0,0,1); glVertex(x2,y1,0)
                glEnd()
                
                if col == self.city.GRASS:
                    # Pave paradise, put up parking lot
                    self.paintBuilding(x1,y1,x2,y2)
                elif col == self.city.SAFEHOUSE:
                    self.paintBuilding(x1,y1,x2,y2,self.SAFEHOUSE)
                colNum += 1
            rowNum += 1
        glEndList()
        return id
            
    def paintBuilding(self,fromX,fromY,toX,toY,type=ORDINARY):
        
        inset = common.BLOCK_UNITS / 20.0
        fromX += inset
        fromY -= inset
        toX -= inset * 2
        toY += inset * 2
        
        glMaterialfv(GL_FRONT_AND_BACK, GL_DIFFUSE, [0.5,0.5,0.5,1.0])
        glMaterialfv(GL_FRONT_AND_BACK, GL_SPECULAR, [0.1, 0.1, 0.1, 1.0])
        
        height = random.randint(20,120)
        
        north = (0,1,0)
        south = (0,-1,0)
        east = (1, 0, 0)
        west = (-1, 0, 0)
        top = (0, 0, 1)
        
        #n/s l/u e/w = north/south, lower/upper, east/west
        nlw = (fromX, fromY, 0)
        nle = (toX, fromY, 0)
        nue = (toX, fromY, height)
        nuw = (fromX, fromY, height)
        
        slw = (fromX, toY, 0)
        sle = (toX, toY, 0)
        sue = (toX, toY, height)
        suw = (fromX, toY, height)
        
        if type == self.ORDINARY:
            texChoices = self.buildTextures
        elif type == self.SAFEHOUSE:
            texChoices = self.safeTextures
        
        tex = random.choice(texChoices)
        glBindTexture(GL_TEXTURE_2D,tex)
        
        glBegin(GL_QUADS)
        glNormal(north)
        glTexCoord2f(1,0); glVertex(nlw)
        glTexCoord2f(1,1); glVertex(nuw)
        glTexCoord2f(0,1); glVertex(nue);
        glTexCoord2f(0,0); glVertex(nle)
        glEnd()
        
        tex = random.choice(texChoices)
        glBindTexture(GL_TEXTURE_2D,tex)
        
        glBegin(GL_QUADS)
        glNormal(south)
        glTexCoord2f(0,0); glVertex(slw)
        glTexCoord2f(1,0); glVertex(sle)
        glTexCoord2f(1,1); glVertex(sue)
        glTexCoord2f(0,1); glVertex(suw);
        glEnd()
        
        tex = random.choice(texChoices)
        glBindTexture(GL_TEXTURE_2D,tex)
        
        glBegin(GL_QUADS)
        glNormal(east)
        glTexCoord2f(0,0); glVertex(sle)
        glTexCoord2f(1,0); glVertex(nle)
        glTexCoord2f(1,1); glVertex(nue)
        glTexCoord2f(0,1); glVertex(sue)
        glEnd()
        
        tex = random.choice(texChoices)
        glBindTexture(GL_TEXTURE_2D,tex)
        
        glBegin(GL_QUADS)
        glNormal(west)
        glTexCoord2f(1,0); glVertex(slw)
        glTexCoord2f(1,1); glVertex(suw)
        glTexCoord2f(0,1); glVertex(nuw)
        glTexCoord2f(0,0); glVertex(nlw);
        glEnd()
        glBindTexture(GL_TEXTURE_2D,0)
        
        if type == self.SAFEHOUSE:
            glMaterialfv(GL_FRONT_AND_BACK, GL_DIFFUSE, [0.7,0.0,0.0,1.0])
            glMaterialfv(GL_FRONT_AND_BACK, GL_SPECULAR, [0.1, 0.1, 0.1, 1.0])
        glBegin(GL_QUADS)
        glNormal(top)
        glVertex(suw); glVertex(sue);
        glVertex(nue); glVertex(nuw);
        glEnd()
        