import math
import types

class Vector:
    
    def __init__(self,x,y,z=0.0,cartesian=True):
        """If cartesian, x and y are x,y.  If not,
           x is magnitude, and y is angle (in degrees)"""
        if cartesian:
            self.x = x
            self.y = y
            self.z = z
        else:
            rads = math.radians(y)
            self.x = x * math.cos(rads)
            self.y = x * math.sin(rads)
            self.z = 0
            
        self.ensureNonZero()
        
    def __add__(self,vector):
        if type(vector) == types.InstanceType:
            return Vector(self.x + vector.x,self.y + vector.y,
                self.z + vector.z)
        else:
            return Vector(self.x + vector,self.y + vector,
                self.z + vector)

    def __div__(self,scale):
        return self*(1.0/scale)

    def __eq__(self,other):
        if other == None: return False
        return self.x == other.x and self.y == other.y and self.z == other.z

    def __getitem__(self,other):
        loc = [self.x,self.y,self.z]
        return loc[other]

    def __len__(self):
        return 3

    def __mul__(self,scale):
        return Vector(self.x * scale, self.y * scale, self.z*scale)

    def __ne__(self,other):
        return not (self == other)

    def __neg__(self):
        return self*-1

    def __sub__(self,vector):
        return self+(-vector)

    def __str__(self):
        return "[%f,%f,%f]" % (self.x,self.y,self.z)
        
    def angle(self):
        return math.atan2(self.y,self.x)
    
    def angleBetween(self, other):
        dot = self.dot(other)
        mag1 = self.magnitude()
        mag2 = other.magnitude()
        return math.acos( dot / (mag1 * mag2) )
        
    def clampMagnitude(self,mag):
        newX,newY,newZ = self.x,self.y,self.z
        if self.magnitude() > mag:
            direction = self.unit()
            direction *= mag
            newX,newY,newZ = direction.x,direction.y,direction.z
        return Vector(newX,newY,newZ)    
        
    def cross(self,other):
        return Vector( self.y*other.z - self.z * other.y,
            -self.x*other.z + self.z * other.x,
            self.x * other.y - other.x * self.y)
        
    def dist(self,other):
        return (self-other).magnitude()
        
    def dot(self,other):
        return self.x * other.x + self.y * other.y + self.z * other.z
        
    def ensureNonZero(self):
        if self.isZero():
            self.x = 0.00001
            self.y = 0.00001
            self.z = 0
        
    def isZero(self):
        return self.x == self.y == self.z == 0
        
    def magnitude(self):
        try:
            return math.sqrt( self.x * self.x + self.y * self.y + self.z * self.z)
        except OverflowError:
            return 4294967296 # closest thing we've got to MAX_INT
    
    def move(self,x,y):
        self.x = x
        self.y = y
    
    def rotate(self,angle):
        x = self.x * math.cos(math.radians(-angle)) + self.y * math.sin(math.radians(-angle));
        y = -self.x * math.sin(math.radians(-angle)) + self.y * math.cos(math.radians(-angle));
        return Vector(x,y,0)
    
    def unit(self):
        mag = self.magnitude()
        if mag == 0:
            return self
        else:
            return Vector(self.x / mag, self.y / mag, self.z/mag)